@@ -98,13 +98,13 @@ class Phase(object):
         self.in_queue = in_queue
         self.out_queue = context.new_queue() if out_queue is True else out_queue
         self.name = name
+        self.context = context
 
         # If we have an in_queue, name it after ourselves
         if in_queue:
             in_queue.name = name
 
         self.__thread = None
-        self.__context = context
         self.__started = False
 
         # future representing the completion of all delayed puts on
@@ -204,22 +204,34 @@ class Phase(object):
           via this function to be resolved.
 
         It's recommended to tune each phase's batching so that this is called
-        no more than 100,000 times in a single phase. Beyond that, scaling
-        issues may occur.
+        no more than 100,000 times in a single phase (and use put_future_outputs
+        instead where possible). Beyond that, scaling issues may occur.
+        """
+
+        # Reuse put_future_outputs wrapping the value as a single-element list.
+        self.put_future_outputs(f_map(value, lambda x: [x]), task_done)
+
+    def put_future_outputs(self, values, task_done=True):
+        """Like put_future_output, but works with a list of values rather
+        than a single value.
+
+        It is more efficient to call this function with a list of size N
+        than to call put_future_output N times.
         """
 
         f = f_map(
-            value,
+            values,
             partial(self.__future_output_done, task_done=task_done),
             error_fn=self.__future_output_failed,
         )
         self.__future_puts = f_and(self.__future_puts, f)
 
-    def __future_output_done(self, value, task_done):
-        # Called when a Future[value] has been resolved successfully.
+    def __future_output_done(self, values, task_done):
+        # Called when a Future[list[value]] has been resolved successfully.
         # Returns True so that the resulting future can be used with f_and
         # to indicate success.
-        self.put_output(value, task_done=task_done)
+        for value in values:
+            self.put_output(value, task_done=task_done)
         return True
 
     def __future_output_failed(self, exception):
@@ -227,7 +239,7 @@ class Phase(object):
 
         # Immediately set context into error state to interrupt all phases
         # (including potentially ourselves) as early as possible.
-        self.__context.set_error()
+        self.context.set_error()
 
         # re-raising the exception will cause it to propagate up through
         # __thread_target, where reasonable logging occurs.
@@ -249,7 +261,7 @@ class Phase(object):
         # OR if the context is in the error state. The latter case
         # avoids us continuing to work on our queue for a long time after
         # we already know there's a fatal error.
-        if out is Phase.ERROR or self.__context.has_error:
+        if out is Phase.ERROR or self.context.has_error:
             raise PhaseInterrupted("Stopping %s due to error" % self.name)
 
         return out
@@ -320,4 +332,4 @@ class Phase(object):
             # Put the context into error state. This will inform all other
             # phases (at least those with an input queue) that we've hit an
             # error, and so they will be interrupted.
-            self.__context.set_error()
+            self.context.set_error()
